import cgi
from datetime import datetime, timedelta

from django.conf.urls import url
from django.contrib.auth import get_user_model
from django.http import HttpResponse
from django.test import TestCase

from urlauth.models import AuthKey
from urlauth.settings import URLAUTH_AUTHKEY_NAME
from urlauth.signals import authkey_processed
from urlauth.util import InvalidKey, load_key

User = get_user_model()


class UrlauthTestCase(TestCase):
    urls = "urlauth.tests"
    test_url = "/urlauth_test_view/"

    # Helpers
    def setUp(self):
        """
        Create active and inactive users.
        """

        User.objects.all().delete()
        self.user = User.objects.create_user("user", "user@host.com", "pass")
        self.ban_user = User.objects.create_user(
            "ban_user", "ban_user@host.com", "pass"
        )
        self.ban_user.is_active = False
        self.ban_user.save()

    def process_url(self, url, **kwargs):
        """
        Add authkey to the ``url`` and then open that url with django test client.

        Returns:
            The ID of created ``AuthKey`` instance
        """

        url = AuthKey.objects.wrap_url(url, **kwargs)
        path, args = url.split("?")[0], cgi.parse_qs(url.split("?")[1])
        self.client.get(url, args)
        return args[URLAUTH_AUTHKEY_NAME][0]

    # Test cases

    def test_create_key(self):
        "Test the create_key function"
        self.assertTrue(
            AuthKey.objects.create_key(uid=self.user.pk, expired=datetime.now())
        )

    def test_wrap_url(self):
        "Test the wrap_url function"
        expired = datetime.now()

        AuthKey.objects.all().delete()
        clean_url = "http://ya.ru"
        url = AuthKey.objects.wrap_url(clean_url, uid=self.user.pk, expired=expired)
        key = AuthKey.objects.get()
        self.assertEqual(
            url, "{}?{}={}".format(clean_url, URLAUTH_AUTHKEY_NAME, key.id)
        )

        AuthKey.objects.all().delete()
        clean_url = "http://ya.ru?foo=bar"
        url = AuthKey.objects.wrap_url(clean_url, uid=self.user.pk, expired=expired)
        key = AuthKey.objects.get()
        self.assertEqual(
            url, "{}&{}={}".format(clean_url, URLAUTH_AUTHKEY_NAME, key.id)
        )

        # Test url with hash
        AuthKey.objects.all().delete()
        clean_url = "http://ya.ru#hash"
        url = AuthKey.objects.wrap_url(clean_url, uid=self.user.pk, expired=expired)
        key = AuthKey.objects.get()
        self.assertEqual(
            url, "http://ya.ru?{}={}#hash".format(URLAUTH_AUTHKEY_NAME, key.id)
        )

        # Test url with hash and non-empty querystring
        AuthKey.objects.all().delete()
        clean_url = "http://ya.ru?foo=bar#hash"
        url = AuthKey.objects.wrap_url(clean_url, uid=self.user.pk, expired=expired)
        key = AuthKey.objects.get()
        self.assertEqual(
            url, "http://ya.ru?foo=bar&{}={}#hash".format(URLAUTH_AUTHKEY_NAME, key.id)
        )

    def test_validate_key(self):
        "Test the validate_key function"
        expired = datetime.now() - timedelta(seconds=1)
        key = AuthKey.objects.create_key(uid=self.user.pk, expired=expired, foo="bar")
        self.assertRaises(InvalidKey, lambda: load_key(key))

        expired = datetime.now() + timedelta(seconds=10)
        key = AuthKey.objects.create_key(uid=self.user.pk, expired=expired)
        self.assertTrue(load_key(key))

    def test_authentication(self):
        "Test the authentication middleware"
        expired = datetime.now() + timedelta(days=1)
        resp = self.client.get(self.test_url)

        # Guest is not authenticated
        self.client.logout()
        self.assertFalse("_auth_user_id" in self.client.session)

        ## Simple authorization
        self.client.logout()
        self.process_url(self.test_url, uid=self.user.pk, expired=expired)
        # self.assertEqual(self.client.session, self.user.pk)

        self.assertEqual(self.client.session["_auth_user_id"], str(self.user.pk))

        ## Baned user can't authorize
        self.client.logout()
        self.process_url(self.test_url, uid=self.ban_user.pk, expired=expired)
        self.assertFalse("_auth_user_id" in self.client.session)

        ## Expired auth key does not work
        self.client.logout()
        expired = datetime.now() - timedelta(seconds=1)
        self.process_url(self.test_url, uid=self.user.pk, expired=expired)
        self.assertFalse("_auth_user_id" in self.client.session)

    def test_signals(self):
        logs = []

        def handler(**k):
            return logs.append("X")

        authkey_processed.connect(handler)

        self.assertEqual(0, len(logs))
        self.client.logout()
        expired = datetime.now() + timedelta(days=1)
        self.process_url(self.test_url, uid=self.user.pk, expired=expired)
        self.assertEqual(1, len(logs))

    def test_onetime_feature(self):
        self.client.logout()
        expired = datetime.now() + timedelta(days=1)
        hash = self.process_url(self.test_url, uid=self.user.pk, expired=expired)
        self.assertFalse(AuthKey.objects.filter(id=hash).count())

        self.client.logout()
        expired = datetime.now() + timedelta(days=1)
        hash = self.process_url(
            self.test_url, uid=self.user.pk, expired=expired, onetime=False
        )
        self.assertTrue(AuthKey.objects.filter(id=hash).count())

    def test_bulk(self):
        User.objects.all().delete()
        AuthKey.objects.all().delete()
        for x in range(100):
            username = "test%d" % x
            User.objects.create_user(username, "%s@gmail.com" % username, username)
        users = User.objects.all()
        for user in users:
            obj = AuthKey.objects.wrap_url("/", uid=user.pk)
        self.assertEqual(User.objects.count(), AuthKey.objects.count())


def test_view(request):
    return HttpResponse("")


urlpatterns = [
    url("urlauth_test_view/", test_view, name="urlauth_test_view"),
]



from django.conf import settings
from django.contrib import auth
from django.contrib.auth import get_user_model
from django.utils.deprecation import MiddlewareMixin

from urlauth.models import AuthKey
from urlauth.settings import URLAUTH_AUTHKEY_NAME
from urlauth.signals import authkey_processed
from urlauth.util import InvalidKey, load_key

User = get_user_model()


class AuthKeyMiddleware(MiddlewareMixin):
    """
    This middleware can authenticate user with an auth key in HTTP request.
    """

    def process_request(self, request):
        key = request.GET.get(
            getattr(settings, "URLAUTH_AUTHKEY_NAME", URLAUTH_AUTHKEY_NAME)
        )
        user = None
        if key is None:
            return
        try:
            key = load_key(key)
            if key.uid:
                try:
                    user = User.objects.get(pk=key.uid)
                except User.DoesNotExist:
                    raise InvalidKey("User [pk=%s] does not exist" % user.pk) from None
        except InvalidKey:
            return

        authkey_processed.send(sender=AuthKey, key=key, request=request, user=user)

        if user and user.is_active:
            user.backend = "django.contrib.auth.backends.ModelBackend"
            auth.login(request, user)

        if key.onetime:
            key.delete()

        request.authkey = key



from datetime import datetime

from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models.sql.subqueries import DeleteQuery

from urlauth.models import AuthKey


def truncate_queryset(qs):
    """
    Deletes all records matched by queryset using

        DELETE from table WHERE <condition>

    query without fetching PK values for all items in original queryset
    unlike django's standard QuerySet.delete().

    See https://code.djangoproject.com/ticket/9519.
    """

    delete_query = qs.query.clone(DeleteQuery)

    # transaction management code is copied from QuerySet.update
    if not transaction.is_managed(using=qs.db):
        transaction.enter_transaction_management(using=qs.db)
        forced_managed = True
    else:
        forced_managed = False
    try:
        delete_query.get_compiler(qs.db).execute_sql(None)
        if forced_managed:
            transaction.commit(using=qs.db)
        else:
            transaction.commit_unless_managed(using=qs.db)
    finally:
        if forced_managed:
            transaction.leave_transaction_management(using=qs.db)


class Command(BaseCommand):
    help = "Remove expired keys from database"

    def handle(self, *args, **kwargs):
        qs = AuthKey.objects.filter(expired__lt=datetime.now())
        count = qs.count()

        # qs.delete() is grossly inefficient for larger querysets
        # and there can be a *LOT* of AuthKey's
        truncate_queryset(qs)

        print("%d keys deleted" % count)



import logging
from datetime import datetime

from urlauth.models import AuthKey


class InvalidKey(Exception):
    pass


def wrap_url(url, **kwargs):
    """
    Create a new authorization key and append it to the url.
    """

    logging.error(
        "urlauth.util.wrap_url is deprecated. Use AuthKey.objects.wrap_url instead."
    )
    return AuthKey.objects.wrap_url(url, **kwargs)


def load_key(hash):
    """
    Load key record associated with given hash.
    """

    try:
        key = AuthKey.objects.get(pk=hash)
    except AuthKey.DoesNotExist:
        raise InvalidKey("Key does not exist")
    if datetime.now() > key.expired:
        raise InvalidKey("Key is expired")
    return key



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-09-16 09:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AuthKey",
            fields=[
                (
                    "id",
                    models.CharField(max_length=40, primary_key=True, serialize=False),
                ),
                ("uid", models.PositiveIntegerField(null=True)),
                ("expired", models.DateTimeField()),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("onetime", models.BooleanField(default=True)),
                ("data", models.TextField()),
            ],
        ),
    ]



import json
from datetime import datetime, timedelta
from uuid import uuid4

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.db.utils import IntegrityError

from urlauth.settings import URLAUTH_AUTHKEY_NAME, URLAUTH_AUTHKEY_TIMEOUT


class URLAuthError(Exception):
    pass


class AuthKeyManager(models.Manager):
    def create_key(self, uid, expired=None, onetime=True, **kwargs):
        """
        Create AuthKey object and return its ID.
        :param uid: The uid of the user to automatically log
        :param expired: (default:  datetime.now() + timedelta(seconds=settings.URLAUTH_AUTHKEY_TIMEOUT)) datetime.now() + time delta to specify when the key must expire
        :param ontime: (default: True) Generate a one time key
        :param kwargs: others arguments
        :return: An unique auth key
        """

        key = AuthKey()
        key.uid = uid
        if expired:
            key.expired = expired
        else:
            key.expired = datetime.now() + timedelta(
                seconds=getattr(
                    settings, "URLAUTH_AUTHKEY_TIMEOUT", URLAUTH_AUTHKEY_TIMEOUT
                )
            )

        key.onetime = onetime
        key.data = json.dumps(kwargs)

        # Try 10 times to create AuthKey instance with unique PK
        for _ in range(10):
            key.pk = uuid4()  # Use a unique identifier
            try:
                key.save(force_insert=True)
            except IntegrityError:
                key.pk = None
            else:
                break

        if not key.pk:
            raise URLAuthError("Could not create unique key")

        return key.id

    def wrap_url(self, url, uid, **kwargs):
        """
        Create new authorization key, append it to the url and return modified url.
        :param url: The url to wrap
        :param uid: id of the user to log
        :param kwargs: others params used byt create_key
        :return: An url wrapped with an unique key
        """

        key_id = self.create_key(uid, **kwargs)
        clue = "?" in url and "&" or "?"
        parts = url.rsplit("#", 1)
        if len(parts) > 1:
            url, hash = parts
            hash = "#%s" % hash
        else:
            hash = ""
        url = "{}{}{}={}{}".format(
            url,
            clue,
            getattr(settings, "URLAUTH_AUTHKEY_NAME", URLAUTH_AUTHKEY_NAME),
            key_id,
            hash,
        )
        return url


class AuthKey(models.Model):
    id = models.CharField(max_length=40, primary_key=True)
    uid = models.PositiveIntegerField(null=True)
    expired = models.DateTimeField()
    created = models.DateTimeField(auto_now_add=True)
    onetime = models.BooleanField(blank=True, default=True)
    data = models.TextField()

    objects = AuthKeyManager()

    def __str__(self):
        return "AuthKey #%s" % self.id

    @property
    def extra(self):
        return json.loads(self.data)

    def get_user(self):
        if self.uid:
            return get_user_model().objects.get(pk=self.uid)
        else:
            return None
